/* Public header.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#ifndef __MPM_H__
#define __MPM_H__

struct __gmpr_struct {
  mp_limb_t  _mp_d[1];
};
typedef struct __gmpr_struct        *mpr_t;
typedef struct __gmpr_struct        *mpr_ptr;
typedef const struct __gmpr_struct  *mpr_srcptr;

typedef struct __gmpm_struct              *mpm_ptr;
typedef __gmp_const struct __gmpm_struct  *mpm_srcptr;

struct __gmpm_struct {
  mpz_t          _mp_m;
  unsigned long  _mp_mbits;        /* number of bits in m */
  mp_size_t      _mp_rsize;        /* limbs used by an mpr_t */
  mp_size_t      _mp_ralloc;       /* limbs allocated to an mpr_t */
  mpr_t          _mp_negone;       /* -1 */
  mp_ptr         _mp_div2ceil;     /* ceil(m/2) */

  mpr_t          _mp_rmod;         /* redc: R mod m */
  mpr_t          _mp_r2mod;        /* redc: R^2 mod m */
  mpr_t          _mp_r3mod;        /* redc: R^3 mod m */
  mp_ptr         _mp_inv;
  mp_limb_t      _mp_smallinv[8];
  mp_ptr         _mp_norm;         /* m normalized (within a NUMB) */
  int            _mp_norm_lshift;  /* left shift for norm */
  int            _mp_norm_rshift;
  mp_limb_t      _mp_norm_rshift_mask;
  mp_limb_t      _mp_norm_dh;      /* norm high, lshift extra NAIL_BITS */
  mp_limb_t      _mp_norm_dl;      /* norm sechigh, lshift extra NAIL_BITS */
  mp_limb_t      _mp_norm_dh_inv;  /* invert_limb of _mp_norm_dh */
  void          *_mp_extension;

  void   (*_mpm_clear)    __GMP_PROTO ((mpm_ptr));

  void   (*_mpr_add)      __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
  void   (*_mpr_add_ui)   __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  void   (*_mpr_addmul_ui)__GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  void   (*_mpr_clear)    __GMP_PROTO ((mpr_ptr, mpm_srcptr));
  void   (*_mpr_div)      __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
  void   (*_mpr_div_2exp) __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  int    (*_mpr_equal)    __GMP_PROTO ((mpr_srcptr, mpr_srcptr, mpm_srcptr));
  int    (*_mpr_equal_negone)  __GMP_PROTO ((mpr_srcptr, mpm_srcptr));
  int    (*_mpr_equal_one)__GMP_PROTO ((mpr_srcptr, mpm_srcptr));
  int    (*_mpr_equal_ui) __GMP_PROTO ((mpr_srcptr, unsigned long, mpm_srcptr));
  int    (*_mpr_equal_zero) __GMP_PROTO ((mpr_srcptr, mpm_srcptr));
  unsigned long (*_mpr_get_ui) __GMP_PROTO ((mpr_srcptr, mpm_srcptr));
  void   (*_mpr_gcd)      __GMP_PROTO ((mpz_ptr, mpr_srcptr, mpm_srcptr));
  void   (*_mpr_get_z)    __GMP_PROTO ((mpz_ptr, mpr_srcptr, mpm_srcptr));
  int    (*_mpr_invert)   __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));
  int    (*_mpr_jacobi)   __GMP_PROTO ((mpr_srcptr, mpm_srcptr));
  void   (*_mpr_mul)      __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
  void   (*_mpr_mul_2exp) __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  void   (*_mpr_mul_si)   __GMP_PROTO ((mpr_ptr, mpr_srcptr, long, mpm_srcptr));
  void   (*_mpr_mul_ui)   __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  void   (*_mpr_neg)      __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));
  mpr_t  (*_mpr_new)      __GMP_PROTO ((mpm_srcptr));
  void   (*_mpr_pow)      __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpz_srcptr, mpm_srcptr));
  void   (*_mpr_pow_ui)   __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  void   (*_mpr_set)      __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));
  void   (*_mpr_set_z)    __GMP_PROTO ((mpr_ptr, mpz_srcptr, mpm_srcptr));
  void   (*_mpr_set_si)   __GMP_PROTO ((mpr_ptr, long,  mpm_srcptr));
  void   (*_mpr_set_ui)   __GMP_PROTO ((mpr_ptr, unsigned long,  mpm_srcptr));

  /* currently unused */
  void   (*_mpr_sqrt)     __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpm_srcptr));

  void   (*_mpr_swap)     __GMP_PROTO ((mpr_ptr, mpr_ptr, mpm_srcptr));
  void   (*_mpr_sub)      __GMP_PROTO ((mpr_ptr, mpr_srcptr, mpr_srcptr, mpm_srcptr));
  void   (*_mpr_sub_ui)   __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  void   (*_mpr_submul_ui)__GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));
  void   (*_mpr_ui_pow)   __GMP_PROTO ((mpr_ptr, unsigned long, mpz_srcptr, mpm_srcptr));
  void   (*_mpr_urandomm) __GMP_PROTO ((mpr_ptr, gmp_randstate_t, mpm_srcptr));

  void   (*_mp_future_1) __GMP_PROTO ((void));
  void   (*_mp_future_2) __GMP_PROTO ((void));
  void   (*_mp_future_3) __GMP_PROTO ((void));
  void   (*_mp_future_4) __GMP_PROTO ((void));
  void   (*_mp_future_5) __GMP_PROTO ((void));
  void   (*_mp_future_6) __GMP_PROTO ((void));
  void   (*_mp_future_7) __GMP_PROTO ((void));
  void   (*_mp_future_8) __GMP_PROTO ((void));
  void   (*_mp_future_9) __GMP_PROTO ((void));
};

typedef struct __gmpm_struct              mpm_t[1];


#define mpr_add(w,x,y,m)      do{ (*(m)->_mpr_add)      (w, x, y, m); }while(0)
#define mpr_add_ui(w,x,y,m)   do{ (*(m)->_mpr_add_ui)   (w, x, y, m); }while(0)
#define mpr_addmul_ui(w,x,y,m)do{ (*(m)->_mpr_addmul_ui)(w, x, y, m); }while(0)
#define mpr_clear(w,m)        do{ (*(m)->_mpr_clear)    (w, m);       }while(0)
#define mpr_div(w,x,y,m)      do{ (*(m)->_mpr_div)      (w, x, y, m); }while(0)
#define mpr_div_2exp(w,x,y,m) do{ (*(m)->_mpr_div_2exp) (w, x, y, m); }while(0)
#define mpr_equal(x,y,m)         ((*(m)->_mpr_equal)    (x, y, m))
#define mpr_equal_negone(x,m)    ((*(m)->_mpr_equal_negone) (x, m))
#define mpr_equal_one(x,m)       ((*(m)->_mpr_equal_one)    (x, m))
#define mpr_equal_ui(x,y,m)      ((*(m)->_mpr_equal_ui) (x, y, m))
#define mpr_equal_zero(x,m)      ((*(m)->_mpr_equal_zero)   (x, m))
#define mpr_gcd(w,x,m)        do{ (*(m)->_mpr_gcd)      (w, x, m);    }while(0)
#define mpr_get_ui(w,x,m)        ((*(m)->_mpr_get_ui)   (x, m))
#define mpr_get_z(w,x,m)      do{ (*(m)->_mpr_get_z)    (w, x, m);    }while(0)
#define mpr_invert(w,x,m)        ((*(m)->_mpr_invert)   (w, x, m))
#define mpr_init(w,m)         do{ (w) = (*(m)->_mpr_new) (m); }while(0)
#define mpr_jacobi(x,m)          ((*(m)->_mpr_jacobi)(x, m))
#define mpr_mul(w,x,y,m)      do{ (*(m)->_mpr_mul)      (w, x, y, m); }while(0)
#define mpr_mul_2exp(w,x,y,m) do{ (*(m)->_mpr_mul_2exp) (w, x, y, m); }while(0)
#define mpr_mul_ui(w,x,y,m)   do{ (*(m)->_mpr_mul_ui)   (w, x, y, m); }while(0)
#define mpr_mul_si(w,x,y,m)   do{ (*(m)->_mpr_mul_si)   (w, x, y, m); }while(0)
#define mpr_neg(w,x,m)        do{ (*(m)->_mpr_neg)      (w, x, m);    }while(0)
#define mpr_set(w,x,m)        do{ (*(m)->_mpr_set)      (w, x, m);    }while(0)
#define mpr_set_z(w,x,m)      do{ (*(m)->_mpr_set_z)    (w, x, m);    }while(0)
#define mpr_set_si(w,x,m)     do{ (*(m)->_mpr_set_si)   (w, x, m);    }while(0)
#define mpr_set_ui(w,x,m)     do{ (*(m)->_mpr_set_ui)   (w, x, m);    }while(0)
#define mpr_swap(x,y,m)       do{ (*(m)->_mpr_swap)     (x, y, m);    }while(0)
#define mpr_pow(w,x,y,m)      do{ (*(m)->_mpr_pow)      (w, x, y, m); }while(0)
#define mpr_sub(w,x,y,m)      do{ (*(m)->_mpr_sub)      (w, x, y, m); }while(0)
#define mpr_sub_ui(w,x,y,m)   do{ (*(m)->_mpr_sub_ui)   (w, x, y, m); }while(0)
#define mpr_submul_ui(w,x,y,m)do{ (*(m)->_mpr_submul_ui)(w, x, y, m); }while(0)
#define mpr_urandomm(w,r,m)   do{ (*(m)->_mpr_urandomm) (w, r, m);    }while(0)

#define mpm_modref(m)   ((m)->_mp_m)

#define mpm_clear  __gmpm_clear
void mpm_clear __GMP_PROTO ((mpm_ptr));

#define mpm_dump  __gmpm_dump
void mpm_dump __GMP_PROTO ((mpm_srcptr));

#define mpm_init  __gmpm_init
void mpm_init __GMP_PROTO ((mpm_ptr, mpz_srcptr));

#define mpm_init_2exp  __gmpm_init_2exp
void mpm_init_2exp __GMP_PROTO ((mpm_ptr, unsigned long));

#define mpm_init_2expsub1  __gmpm_init_2expsub1
void mpm_init_2expsub1 __GMP_PROTO ((mpm_ptr, unsigned long));

#define mpm_init_div  __gmpm_init_div
void mpm_init_div __GMP_PROTO ((mpm_ptr, mpz_srcptr));

#define mpm_init_redc  __gmpm_init_redc
void mpm_init_redc __GMP_PROTO ((mpm_ptr, mpz_srcptr));

#define mpm_init_str  __gmpm_init_str
int mpm_init_str __GMP_PROTO ((mpm_ptr, const char *, int));

#define mpm_init_ui  __gmpm_init_ui
void mpm_init_ui __GMP_PROTO ((mpm_ptr, unsigned long));


#define mpr_dump  __gmpr_dump
void mpr_dump __GMP_PROTO ((mpr_srcptr, mpm_srcptr));

#define mpr_set_str  __gmpr_set_str
int mpr_set_str __GMP_PROTO ((mpr_ptr, const char *, int, mpm_srcptr));

#define mpr_init_set(r,x,m)     \
  do {                          \
    mpr_init (r, m);            \
    mpr_set (r, x, m);          \
  } while (0)

#define mpr_init_set_si(r,x,m)  \
  do {                          \
    mpr_init (r, m);            \
    mpr_set_si (r, x, m);       \
  } while (0)

#define mpr_init_set_ui(r,x,m)  \
  do {                          \
    mpr_init (r, m);            \
    mpr_set_ui (r, x, m);       \
  } while (0)

#define mpr_init_set_z(r,x,m)   \
  do {                          \
    mpr_init (r, m);            \
    mpr_set_z (r, x, m);        \
  } while (0)

#define mpr_init_set_str(r,str,base,m)  \
  ((r) = (*(m)->_mpr_new) (m), mpr_set_str(r,str,base,m))

#define mpr_pow_ui  __gmpr_pow_ui
void mpr_pow_ui __GMP_PROTO ((mpr_ptr, mpr_srcptr, unsigned long, mpm_srcptr));

#define mpr_ui_pow(r, b, e, m)  \
  do {                          \
    mpr_set_ui (r, b, m);       \
    mpr_pow (r, r, e, m);       \
  } while (0)

#define mpr_ui_pow_ui  __gmpr_ui_pow_ui
void mpr_ui_pow_ui (mpr_ptr, unsigned long, unsigned long, mpm_srcptr);

#define mpr_z_pow_ui(r, b, e, m)        \
  do {                                  \
    mpr_set_z (r, b, m);                \
    mpr_pow (r, r, e, m);               \
  } while (0)


int mpr_prp_p (mpr_srcptr, mpm_srcptr);
int mpr_eprp_p (mpr_srcptr, mpm_srcptr);
int mpr_sprp_p (mpr_srcptr, mpm_srcptr);


typedef struct __mpr_lucas_t  *mpr_lucas_ptr;
struct __mpr_lucas_t {
  mpr_t  U, V;
  mpr_t  T, T2;

  mpr_t  Pr, Qr, Dr;
  long   Pl, Ql, Dl;

  mpr_t  two;
  int    jac_DN;

  void (*_mp_double) __GMP_PROTO ((mpr_lucas_ptr, mpm_srcptr));
  void (*_mp_double_vonly) __GMP_PROTO ((mpr_lucas_ptr, mpm_srcptr));
  void (*_mp_step) __GMP_PROTO ((mpr_lucas_ptr, mpm_srcptr));
};
typedef struct __mpr_lucas_t  mpr_lucas_t[1];

void mpr_lucas_init (mpr_lucas_ptr, mpm_srcptr);
void mpr_lucas_clear (mpr_lucas_ptr, mpm_srcptr);

int mpr_lprp_p (mpr_lucas_ptr, mpm_srcptr);
int mpr_elprp_p (mpr_lucas_ptr, mpm_srcptr);
int mpr_slprp_p (mpr_lucas_ptr, mpm_srcptr);


#endif /* __MPM_H__ */
