/* Simple reference implementations of some functions.

   Copyright 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "tests.h"


mp_limb_t
refmpn_addmul_2 (mp_ptr dst, mp_srcptr src, mp_size_t size, mp_srcptr mult)
{
  mp_ptr     src_copy;
  mp_limb_t  c0, c1;

  ASSERT (refmpn_overlap_fullonly_p (dst, src, size));
  ASSERT (! refmpn_overlap_p (dst, size+1, mult, 2));
  ASSERT (size >= 1);
  ASSERT_MPN (mult, 2);

  /* in case dst==src */
  src_copy = refmpn_malloc_limbs (size);
  refmpn_copyi (src_copy, src, size);
  src = src_copy;

  c0 = refmpn_addmul_1 (dst, src, size, mult[0]);
  c1 = refmpn_addmul_1 (dst+1, src, size, mult[1]);
  c1 += ref_addc_limb (&dst[size], dst[size], c0);
  free (src_copy);
  return c1;
}


mp_limb_t
refmpn_addmul_3 (mp_ptr dst, mp_srcptr src, mp_size_t size, mp_srcptr mult)
{
  mp_ptr     src_copy;
  mp_limb_t  c0, c1, c2;

  ASSERT (refmpn_overlap_fullonly_p (dst, src, size));
  ASSERT (! refmpn_overlap_p (dst, size+1, mult, 3));
  ASSERT (size >= 1);
  ASSERT_MPN (mult, 3);

  /* in case dst==src */
  src_copy = refmpn_malloc_limbs (size);
  refmpn_copyi (src_copy, src, size);
  src = src_copy;

  c0 = refmpn_addmul_1 (dst, src, size, mult[0]);
  c1 = refmpn_addmul_1 (dst+1, src, size, mult[1]);
  c2 = refmpn_addmul_1 (dst+1, src, size, mult[2]);

  c1 += ref_addc_limb (&dst[size], dst[size], c0);
  c2 += ref_addc_limb (&dst[size+1], dst[size+1], c1);
  free (src_copy);
  return c1;
}
